/*
 * Copyright © 2005 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <davidr@novell.com>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <X11/Xlib.h>
#include <X11/extensions/Xcomposite.h>

#include <beryl.h>
#include <beryl-private.h>

char *programName;
char **programArgv;
int programArgc;
REGION infiniteRegion;
REGION emptyRegion;

REGION * getInfiniteRegion(void)
{
    return &infiniteRegion;
}

REGION * getEmptyRegion(void)
{
    return &emptyRegion;
}

char * getProgramName(void)
{
    return programName;
}

char ** getProgramArgv(void)
{
    return programArgv;
}

int getProgramArgc(void)
{
    return programArgc;
}

Window currentRoot = 0;

Bool restartSignal = FALSE;

CompWindow *lastFoundWindow = 0;
CompWindow *lastDamagedWindow = 0;

Bool replaceCurrentWm = TRUE;
Bool indirectRendering = FALSE;
Bool strictBinding = FALSE;
Bool useCow = FALSE;
Bool skipGlYield = FALSE;
Bool isNVTFP = FALSE;

static void
usage (void)
{
    printf (_("Usage: %s "
            "[--display DISPLAY] "
            //"[--bg-image PNG]\n             "
#if XCOMPOSITE_MINOR >= 3 || XCOMPOSITE_MAJOR > 0
            "[--use-cow | --no-cow] " 
            "[--strict-binding | --xgl-binding] \n             "
#else
            "[--strict-binding | --xgl-binding] "
#endif
            "[--indirect-rendering | --xgl-rendering]\n             "
            "[--force-nvidia | --force-xgl | --force-aiglx]\n             "
            "[--no-replace | --replace] "
            "[--test-mode]\n             "
            "[--sm-disable] "
            "[--sm-client-id ID] "
            "[--version]\n             "
    	    "[--help] "
            "[PLUGIN]...\n"), programName);
}

static void
signalHandler (int sig)
{
    int status;

    switch (sig)
    {
    case SIGCHLD:
        waitpid (-1, &status, WNOHANG | WUNTRACED);
        break;
    case SIGHUP:
        restartSignal = TRUE;
    default:
        break;
    }
}

static void
detect_nvidia (void)
{
    pid_t p = fork ();
    if (p)
    {
        int stat;
        waitpid (p, &stat, 0);
        if (!WIFEXITED (stat))
        {
            fprintf (stderr,
                     _("Something went wrong with Nvidia detection, giving up."));
            exit (1);
        }
        else
        {
            if (WEXITSTATUS (stat))
            {
                //no Nvidia
                indirectRendering = TRUE;
                useCow = TRUE;
                strictBinding = TRUE;
                printf (_("Nvidia Absent, assuming AIGLX\n"));
            }
            else
            {
                indirectRendering = FALSE;
                useCow = TRUE;
                strictBinding = FALSE;
                isNVTFP=TRUE;
                printf (_("Nvidia Present\n"));
            }
        }
    }
    else
    {
        execl ("/bin/sh", "sh", "-c", "xdpyinfo | grep -q -i NV-GLX",
               (char *)NULL);
        exit (1);
    }

}
static void
detect_xgl (void)
{
    pid_t p = fork ();
    if (p)
    {
        int stat;
        waitpid (p, &stat, 0);
        if (!WIFEXITED (stat))
        {
            fprintf (stderr,
                     _("Something went wrong with Xgl detection, giving up."));
            exit (1);
        }
        else
        {
            if (WEXITSTATUS (stat))
            {
                //no XGL
                printf (_("XGL Absent, checking for NVIDIA\n"));
                detect_nvidia();
            }
            else
            {
                indirectRendering = FALSE;
                useCow = FALSE;
                strictBinding = FALSE;
                printf (_("XGL Present\n"));
            }
        }
    }
    else
    {
        execl ("/bin/sh", "sh", "-c", "xvinfo | grep -q -i Xgl",
               (char *)NULL);
        exit (1);
    }
}

int
main (int argc, char **argv)
{
    char *displayName = 0;
    char *plugin[256];
    int i, nPlugin = 0;
    Bool disableSm = FALSE;
    char *clientId = NULL;
    Bool forceNvidia = FALSE;
    Bool forceXgl = FALSE;
    Bool forceAiglx = FALSE;

    programName = argv[0];
    programArgc = argc;
    programArgv = argv;

    emptyRegion.rects = &emptyRegion.extents;
    emptyRegion.numRects = 0;
    emptyRegion.extents.x1 = 0;
    emptyRegion.extents.y1 = 0;
    emptyRegion.extents.x2 = 0;
    emptyRegion.extents.y2 = 0;
    emptyRegion.size = 0;

    infiniteRegion.rects = &infiniteRegion.extents;
    infiniteRegion.numRects = 1;
    infiniteRegion.extents.x1 = MINSHORT;
    infiniteRegion.extents.y1 = MINSHORT;
    infiniteRegion.extents.x2 = MAXSHORT;
    infiniteRegion.extents.y2 = MAXSHORT;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp (argv[i], "--help"))
        {
            usage ();
            return 0;
        }
        else if (!strcmp (argv[i], "--version"))
        {
            printf (PACKAGE_STRING "\n");
            return 0;
        }
        else if (!strcmp (argv[i], "--display"))
        {
            if (i + 1 < argc)
                displayName = argv[++i];
        }
        else if (!strcmp (argv[i], "--skip-gl-yield"))
        {
            skipGlYield = TRUE;
        }
        else if (!strcmp (argv[i], "--force-nvidia"))
        {
            forceNvidia = TRUE;
            indirectRendering = FALSE;
            useCow = TRUE;
            strictBinding = FALSE;
            isNVTFP = TRUE;
        }
        else if (!strcmp (argv[i], "--force-xgl"))
        {
            forceXgl = TRUE;
            indirectRendering = FALSE;
            useCow = FALSE;
            strictBinding = FALSE;
        }
        else if (!strcmp (argv[i], "--force-aiglx"))
        {
            forceAiglx = TRUE;
            indirectRendering = TRUE;
            useCow = TRUE;
            strictBinding = TRUE;
        }
        else if (!strcmp (argv[i], "--indirect-rendering"))
        {
            indirectRendering = TRUE;
        }
        else if (!strcmp (argv[i], "--xgl-rendering"))
        {
            indirectRendering = FALSE;
        }
        else if (!strcmp (argv[i], "--strict-binding"))
        {
            strictBinding = TRUE;
        }
        else if (!strcmp (argv[i], "--xgl-binding"))
        {
            strictBinding = FALSE;
        }
        else if (!strcmp (argv[i], "--use-cow"))
        {
            useCow = TRUE;
        }
        else if (!strcmp (argv[i], "--no-cow"))
        {
            useCow = FALSE;
        }
        else if (!strcmp (argv[i], "--no-replace"))
        {
            replaceCurrentWm = FALSE;
        }
        else if (!strcmp (argv[i], "--replace"))
        {
            replaceCurrentWm = TRUE;
        }
        else if (!strcmp (argv[i], "--sm-disable"))
        {
            disableSm = TRUE;
        }
        else if (!strcmp (argv[i], "--sm-client-id"))
        {
            if (i + 1 < argc)
                clientId = argv[++i];
        }
        else
        {
            if (nPlugin < 256)
                plugin[nPlugin++] = argv[i];
        }
    }

    if (!forceNvidia && !forceXgl && !forceAiglx)
    {
        detect_xgl ();
    }

    if (isNVTFP && !skipGlYield)
    {
        char * newargs[argc+2];
        int i;
        printf(_("Relaunching %s with __GL_YIELD=\"NOTHING\"\n"),argv[0]);
        setenv("__GL_YIELD","NOTHING",1);
        for (i=0;i<argc;i++)
            newargs[i]=argv[i];
        newargs[argc]="--skip-gl-yield";
        newargs[argc+1]=NULL;
        execvp(argv[0],newargs);
    }

    /*if (!nPlugin)
    {
        nPlugin = 1;
        plugin[0] = "settings";
    }*/

    signal (SIGHUP, signalHandler);
    signal (SIGCHLD, signalHandler);

    if (!disableSm)
        initSession (clientId);

    if (!addDisplay (displayName, plugin, nPlugin))
        return 1;

    eventLoop ();

    if (!disableSm)
        closeSession ();

    return 0;
}
